package IkiWiki::Plugin::table;

use warnings;
use strict;
use Carp;

use utf8;
use English;

use IkiWiki '1.00';
use IkiWiki::Plugin::mdwn;
use Text::CSV;

our $VERSION = '0.6';

our %_defaults = (
    data            =>  undef,      
    file            =>  undef,
    format          =>  'auto',
    sep_char       =>  {
        'csv'   =>  ',',
        'dsv'   =>  '\|',
        },
    class           =>  undef,
    caption         =>  undef,
    no_header       =>  0,
    _has_thead      =>  0,
);
                            
# Module implementation here
sub import { #{{{
        hook(type => "preprocess", id => "table", call => \&preprocess);
} # }}}

sub preprocess (@) { #{{{
    my %params = (%_defaults, @_);

    # Adjust the default identifier 
    if (defined($params{delimiter})) {
        $params{sep_char}->{$params{format}} = $params{delimiter};
    }

    # Load and parse the data 
    my $data = _read_data(%params);

    # we have any data ? 
    if (not scalar(@{$data->{body}})) {
        return "[[table empty data ]]";
    }
    else {
        # Add the html header and the thead row ...
        my @html_lines = _open_table( \%params, $data );

        # Add the body lines 
        push(@html_lines, _build_rows( \%params, $data ));

        # And the html footer ...
        push(@html_lines, _close_table( \%params, $data ));

        # Add indentation             
        my $html = _tidy_up(@html_lines);

        # If the data are in a page ...
        if (defined($params{file})) {
            return sprintf("%s\n\n[[%s|%s]]\n", $html,
                    gettext('Direct data download'),
                    $params{file});
        }
        else {  
            return $html;
        }            
    }
} # }}}

sub _tidy_up {
    my  @lines          =   @_;
    my  $html           =   undef;

    foreach my $text (@lines) {
        my $indentation = $text =~ m{thead>|tbody>}   ? 0 :
                          $text =~ m{tr>}             ? 4 :
                          $text =~ m{td>|th>}         ? 8 :
                                                        0;

        $html .= sprintf("%s%s\n", ' ' x $indentation, $text );
    }        

    return $html;
}    

#
#   this function returns a hash reference with the parsed data:
#       - head: array reference with the header fields
#       - body: array of arrays reference with the rows and body fields
#

sub _read_data {
    my  %params     =   @_;
    my  %data = (
            head    =>  [],
            body    =>  [],
        );
    my  $raw_data   =   $params{data};

    # read the page contents if we have one
    if (defined($params{file})) {
        $raw_data = readfile(srcfile($params{file}));
    }

    # and try to determine the format ...
    if (lc($params{format}) eq 'auto') {
        _determine_format( \%params, $raw_data );
    }

    # and split in lines 
    my  @data_lines = split(/\n/, $raw_data);

    #   special call for csv format
    if (lc($params{format}) eq 'csv') {
        return _read_csv(\%params, \%data, @data_lines);
    }
    #   special call for dsv format
    elsif (lc($params{format}) eq 'dsv') {
        return _read_dsv(\%params, \%data, @data_lines);
    }
    #   fatal error 
    else {
        error(sprintf(gettext('unknow data format %s in table plugin'),
                    $params{format}));
    }
}

sub _determine_format {
    my  $params_ref     =   shift;
    my  $raw_data       =   shift;

    # first try with the more simple format ...        
    if (_is_dsv_data($raw_data)) {
        $params_ref->{format} = 'dsv';
        $params_ref->{sep_char}->{dsv} = '\|';
    }
    #  falling to the default values ...
    else {
        $params_ref->{format} = 'csv';
        $params_ref->{sep_char}->{csv} = ',';
    }

    return;
}

sub _is_dsv_data {
    my  $text       =   shift;

    # get the first line 
    my ($line) = split(/\n/, $text);

    # and search for the standard delimiter
    return $line =~ m{.+\|};
}

sub _read_csv {
    my  $params_ref     =   shift;
    my  $data_ref       =   shift;
    my  @text_lines     =   @_;

    #   build a parser object
    my  $csv =  Text::CSV->new( { 
                    sep_char    =>  $params_ref->{sep_char}->{'csv'},
                    binary      =>  1,
                } ) || error(gettext('could not create a Text::CSV object'));

    #   read lines of raw data ...
    my  $line_number    =   1;
    foreach my $line (@text_lines) {
        # and parse in fields ...
        if ($csv->parse($line)) {
            push(@{$data_ref->{body}}, [ $csv->fields() ]);
        }
        else {
            debug(sprintf(gettext('parse fail at line %d: %s'), 
                    $line_number, $csv->error_input()));
        }
        $line_number++;
    }

    _extract_header($params_ref, $data_ref);

    return $data_ref;
}   

sub _read_dsv {
    my  $params_ref     =   shift;
    my  $data_ref       =   shift;
    my  @text_lines     =   @_;

    my  $splitter = qr{$params_ref->{sep_char}->{'dsv'}};
    foreach my $line (@text_lines) {
        my @fields = split( $splitter, $line );
        push(@{$data_ref->{body}}, [ @fields ]);
    }
    
    _extract_header( $params_ref, $data_ref );

    return $data_ref;
}

sub _extract_header {
    my  $params_ref     =   shift;
    my  $data_ref       =   shift;

    #   is the first line the header line ?
    if ($params_ref->{no_header} == 0) {
        # move the first line from data to header, if we have any cells
        $data_ref->{head} = shift @{$data_ref->{body}}
            if scalar(@{$data_ref->{body}});
    }
    
    return;
}

sub _build_rows {
    my  $params_ref =   shift;
    my  $data_ref   =   shift;
    my  @items      =   ();

    #   Loop around every data record ...
    foreach my $record (@{$data_ref->{body}}) {
        push(@items,'<tr>');

        # Loop around every field ...
        foreach my $field (@{ $record }) {
            # preprocess the content of the cell 
            my $text = _htmlize( $params_ref, $field );

            # Add between html marks                                            
            push(@items,sprintf('<td>%s</td>', $text));
        }
        
        push(@items,'</tr>');
    }        

    return @items;
}

sub _htmlize {
    my  $params_ref =   shift;
    my  $text       =   shift;

    # first preprocess the content of cell
    $text = IkiWiki::preprocess($params_ref->{page}, 
                                $params_ref->{destpage},
                                $text);

    # and second convert to html (including typographic marks) using the
    # markdown plugin directly
    $text = IkiWiki::htmlize( $params_ref->{page}, q(mdwn), $text );

    # cleaning line ends
    chomp $text;

    return $text;
}

sub _open_table {
    my  $params_ref =   shift;
    my  $data_ref   =   shift;
    my  @items      =   ();

    # is there a table class ? 
    if (defined($params_ref->{class})) {
        push(@items, sprintf('<table class="%s">', $params_ref->{class}));
    }
    else {
        push(@items, '<table>');
    }

    # are there any head cells ?
    if (scalar(@{$data_ref->{head}})) {
        push(@items, '<thead>','<tr>');
        foreach my $cell (@{$data_ref->{head}}) {
            my $text = _htmlize( $params_ref, $cell );

            push(@items,sprintf('<th>%s</th>', $text ));
        }
        push(@items, '</tr>','</thead>','<tbody>');
        $params_ref->{_has_thead} = 1;
    }

    return @items;
}
                                    
sub _close_table {
    my  $params_ref =   shift;
    my  @items      =   ();

    if ($params_ref->{_has_thead}) {
        push(@items,'</tbody>');
    }

    push(@items, '</table>');

    return @items;
}

1; # Magic true value required at end of module
__END__

=head1 NAME

IkiWiki::Plugin::table - Build HTML tables from CSV or DSV data

=head1 SYNOPSIS

In any source page include the following:

    The next table show the results:

    [[table class="myclass" format=dsv data="""
    Custom|Amount|
    Fulanito|134,34|
    Menganito|234,56|
    """]]

    This is my last acquisition:

    [[table class="book_record" format=csv file="data/books/record1"]]

    Also you can download [[the book record|data/books/record1]] in CSV format.

And the record1 page should be similar to:

    "Title","Perl Best Practices"
    "Author","Damian Conway"
    "Publisher","O'Reilly"

=head1 DESCRIPTION

This ikiwiki plugin build html tables from comma-separated values. The data can
come from the I<data> parameter or from a ikiwiki page. 

The plugin accept the following parameters:

=over 

=item data | file (required)

The data parameter contains the values for the table. The file parameter is the
ikiwiki page which contains the data. 

=item format (optional)

Is the format of data source. It can be CSV (comma-separated values) or DSV (delimiter-separated values). 

The default value is I<CSV>.

=item delimiter (optional)

Define the character used for separated the fields. By default is the vertical
barr C<|> in DSV format and the comma C<,> for the CSV format.

=item class (optional)

CSS class for the table html element

=item caption (optional)

Text string with the table caption.

=item no_header (optional)

This switch disable the generation of table header elements. By default this
switch is enabled, and the header cells come from the firts data line.

=back

=head1 DEPENDENCIES

=over 

=item L<Text::CSV>

=back

=head1 BUGS AND LIMITATIONS

No bugs have been reported.

Please report any bugs or feature requests to the author. Patches and comments
are welcome.

=head1 AUTHOR

"Vctor Moral"  C<< <victor@taquiones.net> >>

=head1 LICENCE AND COPYRIGHT

Copyright (c) 2006, "Vctor Moral".

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

